import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_add():
    input1 = cle.push(np.asarray([1, 2, 3]))
    input2 = cle.push(np.asarray([4, 5, 6]))
    reference = cle.push(np.asarray([5, 7, 9]))

    output = input1 + input2
    result = cle.pull(output)

    print(result)

    assert np.array_equal(result, reference)


def test_add_with_scalar():
    input1 = cle.push(np.asarray([[1, 2, 3]]))
    input2 = 5
    reference = cle.push(np.asarray([[6, 7, 8]]))

    output = input1 + input2
    result = cle.pull(output)

    print(result)

    assert np.array_equal(result, reference)


def test_add_with_np():
    input1 = cle.push(np.asarray([[4, 2, -8]]))
    input2 = np.asarray([[2, 2, 2]])
    reference = cle.push(np.asarray([[6, 4, -6]]))

    output = input1 + input2
    result = cle.pull(output)

    print(result)

    assert np.array_equal(result, reference)


def test_iadd():
    input1 = cle.push(np.asarray([[1, 2, 3]]))
    input2 = cle.push(np.asarray([[4, 5, 6]]))
    reference = cle.push(np.asarray([[5, 7, 9]]))

    input1 += input2
    result = cle.pull(input1)

    print(result)

    assert np.array_equal(result, reference)


def test_iadd_with_scalar():
    input1 = cle.push(np.asarray([[1, 2, 3]]))
    input2 = 5
    reference = cle.push(np.asarray([[6, 7, 8]]))

    input1 += input2
    result = cle.pull(input1)

    print(result)

    assert np.array_equal(result, reference)


def test_iadd_with_np():
    input1 = cle.push(np.asarray([[4, 2, -8]]))
    input2 = np.asarray([[2, 2, 2]])
    reference = cle.push(np.asarray([[6, 4, -6]]))

    input1 += input2
    result = cle.pull(input1)

    print(result)

    assert np.array_equal(result, reference)


def test_subtract():
    input1 = cle.push(np.asarray([[4, 2, 3]]))
    input2 = cle.push(np.asarray([[1, 5, 6]]))
    reference = cle.push(np.asarray([[3, -3, -3]]))

    output = input1 - input2
    result = cle.pull(output)

    print(result)

    assert np.array_equal(result, reference)


def test_subtract_with_scalar():
    input1 = cle.push(np.asarray([[4, 2, 3]]))
    input2 = 5
    reference = cle.push(np.asarray([[-1, -3, -2]]))

    output = input1 - input2
    result = cle.pull(output)

    print(result)

    assert np.array_equal(result, reference)


def test_subtract_with_np():
    input1 = cle.push(np.asarray([[4, 2, 3]]))
    input2 = np.asarray([[1, 5, 6]])
    reference = cle.push(np.asarray([[3, -3, -3]]))

    output = input1 - input2
    result = cle.pull(output)

    print(result)

    assert np.array_equal(result, reference)


def test_isubtract():
    input1 = cle.push(np.asarray([[4, 2, 3]]))
    input2 = cle.push(np.asarray([[1, 5, 6]]))
    reference = cle.push(np.asarray([[3, -3, -3]]))

    input1 -= input2
    result = cle.pull(input1)

    print(result)

    assert np.array_equal(result, reference)


def test_isubtract_with_scalar():
    input1 = cle.push(np.asarray([[4, 2, 3]]))
    input2 = 5
    reference = cle.push(np.asarray([[-1, -3, -2]]))

    input1 -= input2
    result = cle.pull(input1)

    print(result)

    assert np.array_equal(result, reference)


def test_isubtract_with_np():
    input1 = cle.push(np.asarray([[4, 2, 3]]))
    input2 = np.asarray([[1, 5, 6]])
    reference = cle.push(np.asarray([[3, -3, -3]]))

    input1 -= input2
    result = cle.pull(input1)

    print(result)

    assert np.array_equal(result, reference)


def test_divide():
    input1 = cle.push(np.asarray([[4, 2, -8]]))
    input2 = cle.push(np.asarray([[2, 2, 2]]))
    reference = cle.push(np.asarray([[2, 1, -4]]))

    output = input1 / input2
    result = cle.pull(output)

    print(result)

    assert np.array_equal(result, reference)


def test_divide_with_scalar():
    input1 = cle.push(np.asarray([[4, 2, -8]]))
    input2 = 2
    reference = cle.push(np.asarray([[2, 1, -4]]))

    output = input1 / input2
    result = cle.pull(output)

    print(result)

    assert np.array_equal(result, reference)


def test_divide_with_np():
    input1 = cle.push(np.asarray([[4, 2, -8]]))
    input2 = np.asarray([[2, 2, 2]])
    reference = cle.push(np.asarray([[2, 1, -4]]))

    output = input1 / input2
    result = cle.pull(output)

    print(result)

    assert np.array_equal(result, reference)


def test_idivide():
    input1 = cle.push(np.asarray([[4, 2, -8]]))
    input2 = cle.push(np.asarray([[2, 2, 2]]))
    reference = cle.push(np.asarray([[2, 1, -4]]))

    input1 /= input2
    result = cle.pull(input1)

    print(result)

    assert np.array_equal(result, reference)


def test_idivide_with_scalar():
    input1 = cle.push(np.asarray([[4, 2, -8]]))
    input2 = 2
    reference = cle.push(np.asarray([[2, 1, -4]]))

    input1 /= input2
    result = cle.pull(input1)

    print(result)

    assert np.array_equal(result, reference)


def test_idivide_with_np():
    input1 = cle.push(np.asarray([[4, 2, -8]]))
    input2 = np.asarray([[2, 2, 2]])
    reference = cle.push(np.asarray([[2, 1, -4]]))

    input1 /= input2
    result = cle.pull(input1)

    print(result)

    assert np.array_equal(result, reference)


def test_multiply():
    input1 = cle.push(np.asarray([[4, 2, -8]]))
    input2 = cle.push(np.asarray([[2, 2, 2]]))
    reference = cle.push(np.asarray([[8, 4, -16]]))

    output = input1 * input2
    result = cle.pull(output)

    print(result)

    assert np.array_equal(result, reference)


def test_multiply_with_scalar():
    input1 = cle.push(np.asarray([[4, 2, -8]]))
    input2 = 2
    reference = cle.push(np.asarray([[8, 4, -16]]))

    output = input1 * input2
    result = cle.pull(output)

    print(result)

    assert np.array_equal(result, reference)


def test_multiply_with_np():
    input1 = cle.push(np.asarray([[4, 2, -8]]))
    input2 = np.asarray([[2, 2, 2]])
    reference = cle.push(np.asarray([[8, 4, -16]]))

    output = input1 * input2
    result = cle.pull(output)

    print(result)

    assert np.array_equal(result, reference)


def test_imultiply():
    input1 = cle.push(np.asarray([[4, 2, -8]]))
    input2 = cle.push(np.asarray([[2, 2, 2]]))
    reference = cle.push(np.asarray([[8, 4, -16]]))

    input1 *= input2
    result = cle.pull(input1)

    print(result)

    assert np.array_equal(result, reference)


def test_imultiply_with_scalar():
    input1 = cle.push(np.asarray([[4, 2, -8]]))
    input2 = 2
    reference = cle.push(np.asarray([[8, 4, -16]]))

    input1 *= input2
    result = cle.pull(input1)

    print(result)

    assert np.array_equal(result, reference)


def test_imultiply_with_np():
    input1 = cle.push(np.asarray([[4, 2, -8]]))
    input2 = np.asarray([[2, 2, 2]])
    reference = cle.push(np.asarray([[8, 4, -16]]))

    input1 *= input2
    result = cle.pull(input1)

    print(result)

    assert np.array_equal(result, reference)


def test_gt():
    input1 = cle.push(np.asarray([[4, 2, -8]]))
    input2 = cle.push(np.asarray([[2, 2, 2]]))
    reference = cle.push(np.asarray([[1, 0, 0]]))

    output = input1 > input2
    result = cle.pull(output)

    print(result)

    assert np.array_equal(result, reference)


def test_gt_with_scalar():
    input1 = cle.push(np.asarray([[4, 2, -8]]))
    input2 = 2
    reference = cle.push(np.asarray([[1, 0, 0]]))

    output = input1 > input2
    result = cle.pull(output)

    print(result)

    assert np.array_equal(result, reference)


def test_gt_with_np():
    input1 = cle.push(np.asarray([[4, 2, -8]]))
    input2 = np.asarray([[2, 2, 2]])
    reference = cle.push(np.asarray([[1, 0, 0]]))

    output = input1 > input2
    result = cle.pull(output)

    print(result)

    assert np.array_equal(result, reference)


def test_ge():
    input1 = cle.push(np.asarray([[4, 2, -8]]))
    input2 = cle.push(np.asarray([[2, 2, 2]]))
    reference = cle.push(np.asarray([[1, 1, 0]]))

    output = input1 >= input2
    result = cle.pull(output)

    print(result)

    assert np.array_equal(result, reference)


def test_ge_with_scalar():
    input1 = cle.push(np.asarray([[4, 2, -8]]))
    input2 = 2
    reference = cle.push(np.asarray([[1, 1, 0]]))

    output = input1 >= input2
    result = cle.pull(output)

    print(result)

    assert np.array_equal(result, reference)


def test_ge_with_np():
    input1 = cle.push(np.asarray([[4, 2, -8]]))
    input2 = np.asarray([[2, 2, 2]])
    reference = cle.push(np.asarray([[1, 1, 0]]))

    output = input1 >= input2
    result = cle.pull(output)

    print(result)

    assert np.array_equal(result, reference)


def test_lt():
    input1 = cle.push(np.asarray([[4, 2, -8]]))
    input2 = cle.push(np.asarray([[2, 2, 2]]))
    reference = cle.push(np.asarray([[0, 0, 1]]))

    output = input1 < input2
    result = cle.pull(output)

    print(result)

    assert np.array_equal(result, reference)


def test_lt_with_scalar():
    input1 = cle.push(np.asarray([[4, 2, -8]]))
    input2 = 2
    reference = cle.push(np.asarray([[0, 0, 1]]))

    output = input1 < input2
    result = cle.pull(output)

    print(result)

    assert np.array_equal(result, reference)


def test_lt_with_np():
    input1 = cle.push(np.asarray([[4, 2, -8]]))
    input2 = np.asarray([[2, 2, 2]])
    reference = cle.push(np.asarray([[0, 0, 1]]))

    output = input1 < input2
    result = cle.pull(output)

    print(result)

    assert np.array_equal(result, reference)


def test_le():
    input1 = cle.push(np.asarray([[4, 2, -8]]))
    input2 = cle.push(np.asarray([[2, 2, 2]]))
    reference = cle.push(np.asarray([[0, 1, 1]]))

    output = input1 <= input2
    result = cle.pull(output)

    print(result)

    assert np.array_equal(result, reference)


def test_le_with_scalar():
    input1 = cle.push(np.asarray([[4, 2, -8]]))
    input2 = 2
    reference = cle.push(np.asarray([[0, 1, 1]]))

    output = input1 <= input2
    result = cle.pull(output)

    print(result)

    assert np.array_equal(result, reference)


def test_le_with_np():
    input1 = cle.push(np.asarray([[4, 2, -8]]))
    input2 = np.asarray([[2, 2, 2]])
    reference = cle.push(np.asarray([[0, 1, 1]]))

    output = input1 <= input2
    result = cle.pull(output)

    print(result)

    assert np.array_equal(result, reference)


def test_eq():
    input1 = cle.push(np.asarray([[4, 2, -8]]))
    input2 = cle.push(np.asarray([[2, 2, 2]]))
    reference = cle.push(np.asarray([[0, 1, 0]]))

    output = input1 == input2
    result = cle.pull(output)

    print(result)

    assert np.array_equal(result, reference)


def test_eq_with_scalar():
    input1 = cle.push(np.asarray([[4, 2, -8]]))
    input2 = 2
    reference = cle.push(np.asarray([[0, 1, 0]]))

    output = input1 == input2
    result = cle.pull(output)

    print(result)

    assert np.array_equal(result, reference)


def test_eq_with_np():
    input1 = cle.push(np.asarray([[4, 2, -8]]))
    input2 = np.asarray([[2, 2, 2]])
    reference = cle.push(np.asarray([[0, 1, 0]]))

    output = input1 == input2
    result = cle.pull(output)

    print(result)

    assert np.array_equal(result, reference)


def test_ne():
    input1 = cle.push(np.asarray([[4, 2, -8]]))
    input2 = cle.push(np.asarray([[2, 2, 2]]))
    reference = cle.push(np.asarray([[1, 0, 1]]))

    output = input1 != input2
    result = cle.pull(output)

    print(result)

    assert np.array_equal(result, reference)


def test_ne_with_scalar():
    input1 = cle.push(np.asarray([[4, 2, -8]]))
    input2 = 2
    reference = cle.push(np.asarray([[1, 0, 1]]))

    output = input1 != input2
    result = cle.pull(output)

    print(result)

    assert np.array_equal(result, reference)


def test_ne_with_np():
    input1 = cle.push(np.asarray([[4, 2, -8]]))
    input2 = np.asarray([[2, 2, 2]])
    reference = cle.push(np.asarray([[1, 0, 1]]))

    output = input1 != input2
    result = cle.pull(output)

    print(result)

    assert np.array_equal(result, reference)


def test_pos():
    input1 = cle.push(np.asarray([[4, 2, -8]]))
    reference = cle.push(np.asarray([[4, 2, -8]]))

    output = +input1
    result = cle.pull(output)

    print(result)

    assert np.array_equal(result, reference)


def test_neg():
    input1 = cle.push(np.asarray([[4, 2, -8]]))
    reference = cle.push(np.asarray([[-4, -2, 8]]))

    output = -input1
    result = cle.pull(output)

    print(result)

    assert np.array_equal(result, reference)


def test_power():
    input1 = cle.push(np.asarray([[4, 2, -8]]))
    input2 = cle.push(np.asarray([[2, 2, 2]]))
    reference = cle.push(np.asarray([[16, 4, 64]]))

    output = input1**input2
    result = cle.pull(output)

    print(result)

    assert np.array_equal(result, reference)


def test_power_with_scalar():
    input1 = cle.push(np.asarray([[4, 2, -8]]))
    input2 = 2
    reference = cle.push(np.asarray([[16, 4, 64]]))

    output = input1**input2
    result = cle.pull(output)

    print(result)

    assert np.array_equal(result, reference)


def test_power_with_np():
    input1 = cle.push(np.asarray([[4, 2, -8]]))
    input2 = np.asarray([[2, 2, 2]])
    reference = cle.push(np.asarray([[16, 4, 64]]))

    output = input1**input2
    result = cle.pull(output)

    print(result)

    assert np.array_equal(result, reference)


def test_ipower():
    input1 = cle.push(np.asarray([[4, 2, -8]]))
    input2 = cle.push(np.asarray([[2, 2, 2]]))
    reference = cle.push(np.asarray([[16, 4, 64]]))

    input1 **= input2
    result = cle.pull(input1)

    print(result)

    assert np.array_equal(result, reference)


def test_ipower_with_scalar():
    input1 = cle.push(np.asarray([[4, 2, -8]]))
    input2 = 2
    reference = cle.push(np.asarray([[16, 4, 64]]))

    input1 **= input2
    result = cle.pull(input1)

    print(result)

    assert np.array_equal(result, reference)


def test_ipower_with_np():
    input1 = cle.push(np.asarray([[4, 2, -8]]))
    input2 = np.asarray([[2, 2, 2]])
    reference = cle.push(np.asarray([[16, 4, 64]]))

    input1 **= input2
    result = cle.pull(input1)

    print(result)

    assert np.array_equal(result, reference)


def test_ipow_with_types():
    scalars = [
        2,
        int(2),
        float(2),
        np.asarray([[2]]).astype(np.int8).max(),
        np.asarray([[2]]).astype(np.uint8).max(),
        np.asarray([[2]]).astype(np.int16).max(),
        np.asarray([[2]]).astype(np.uint16).max(),
        np.asarray([[2]]).astype(np.int32).max(),
        np.asarray([[2]]).astype(np.uint32).max(),
        np.asarray([[2]]).astype(np.int64).max(),
        np.asarray([[2]]).astype(np.uint64).max(),
        np.asarray([[2]]).astype(np.float32).max(),
        np.asarray([[2]]).astype(np.float64).max(),
        2,
    ]

    for input2 in scalars:
        print("Testing type", type(input2))

        input1 = cle.push(np.asarray([[4, 2, -8]]))
        reference = cle.push(np.asarray([[16, 4, 64]]))

        input1 **= input2
        result = cle.pull(input1)

        print(result)

        assert np.array_equal(result, reference)



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_minimum_of_all_pixels():
    np_input = np.asarray(
        [
            [[1, 2, 3, 10], [4, 5, 6, 11], [7, 8, 9, 12]],
            [[1, 2, 3, 13], [4, 5, 6, 14], [7, 8, 9, 15]],
        ]
    )

    gpu_input = cle.push(np_input)
    result = cle.minimum_of_all_pixels(gpu_input)
    print(result)
    assert result == 1

    gpu_input = cle.push(np_input)
    result = cle.minimum_of_all_pixels(gpu_input)
    print(result)
    assert result == 1



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


input1 = np.asarray([1, 2, 3])
input2 = np.asarray([4, 5, 7])


def test_subtract_images():
    reference = np.asarray([3, 3, 4])
    output = cle.subtract_images(input2, input1)
    result = cle.pull(output)

    print(output)
    print(result)

    assert np.array_equal(result, reference)



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_square():
    test1 = cle.push(np.asarray([4, 6]))

    reference = cle.push(np.asarray([16, 36]))

    result = cle.square(test1)

    print(result)

    a = cle.pull(result)
    b = cle.pull(reference)

    print(a)
    print(b)

    assert np.allclose(a, b, 0.001)



#
# This code is auto-generated from CLIc 'cle::tier6.hpp' file, do not edit manually.
#

import importlib
import warnings
from typing import Optional

import numpy as np

from ._array import Image
from ._core import Device
from ._decorators import plugin_function

clic = importlib.import_module('._pyclesperanto', package='pyclesperanto')

@plugin_function(categories=["label processing", "in assistant", "bia-bob-suggestion"])
def dilate_labels(
    input_image: Image,
    output_image: Optional[Image] =None,
    radius: int =2,
    device: Optional[Device] =None
) -> Image:
    """Dilates labels to a larger size. No label overwrites another label. Similar to
    the implementation in scikitimage [2] and MorpholibJ[3] Notes * This operation
    assumes input images are isotropic.

    Parameters
    ----------
    input_image: Image 
        label image to erode
    output_image: Optional[Image] (= None)
        result
    radius: int (= 2)
        
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image
    """
    return clic._dilate_labels(device, input_image, output_image, int(radius))

@plugin_function(categories=["label processing", "in assistant"])
def erode_labels(
    input_image: Image,
    output_image: Optional[Image] =None,
    radius: int =1,
    relabel: bool =False,
    device: Optional[Device] =None
) -> Image:
    """Erodes labels to a smaller size. Note: Depending on the label image and the
    radius, labels may disappear and labels may split into multiple islands. Thus,
    overlapping labels of input and output may not have the same identifier. Notes *
    This operation assumes input images are isotropic.

    Parameters
    ----------
    input_image: Image 
        result
    output_image: Optional[Image] (= None)
        
    radius: int (= 1)
        
    relabel: bool (= False)
        and all label indices exist.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image
    """
    return clic._erode_labels(device, input_image, output_image, int(radius), relabel)

@plugin_function(categories=["label", "in assistant"])
def gauss_otsu_labeling(
    input_image0: Image,
    output_image: Optional[Image] =None,
    outline_sigma: float =0,
    device: Optional[Device] =None
) -> Image:
    """Labels objects directly from grey-value images.  The outline_sigma parameter
    allows tuning the segmentation result. Under the hood,  this filter applies a
    Gaussian blur, Otsu-thresholding [1] and connected component labeling [2]. The
    thresholded binary image is flooded using the Voronoi tesselation approach
    starting from the found local maxima.

    Parameters
    ----------
    input_image0: Image 
        intensity image to add labels
    output_image: Optional[Image] (= None)
        Output label image.
    outline_sigma: float (= 0)
        Gaussian blur sigma along all axes
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://ieeexplore.ieee.org/document/4310076
    [2] https://en.wikipedia.org/wiki/Connected-component_labeling
    """
    return clic._gauss_otsu_labeling(device, input_image0, output_image, float(outline_sigma))

@plugin_function(categories=["label"])
def masked_voronoi_labeling(
    input_image: Image,
    mask: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Takes a binary image, labels connected components and dilates the regions using
    a octagon shape until they touch. The region growing is limited to a masked
    area. The resulting label map is written to the output.

    Parameters
    ----------
    input_image: Image 
        
    mask: Image 
        
    output_image: Optional[Image] (= None)
        
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_maskedVoronoiLabeling
    """
    return clic._masked_voronoi_labeling(device, input_image, mask, output_image)

@plugin_function(categories=["label", "in assistant", "bia-bob-suggestion"])
def voronoi_labeling(
    input_image: Image,
    output_image: Optional[Image] =None,
    device: Optional[Device] =None
) -> Image:
    """Takes a binary image, labels connected components and dilates the regions using
    a octagon shape until they touch. The resulting label map is written to the
    output.

    Parameters
    ----------
    input_image: Image 
        
    output_image: Optional[Image] (= None)
        
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_voronoiLabeling
    """
    return clic._voronoi_labeling(device, input_image, output_image)

@plugin_function(categories=["label processing", "in assistant", "bia-bob-suggestion"])
def remove_small_labels(
    input_image: Image,
    output_image: Optional[Image] =None,
    min_size: float =100,
    device: Optional[Device] =None
) -> Image:
    """Removes labelled objects small than a given size (in pixels) from a label map.

    Parameters
    ----------
    input_image: Image 
        Label image to filter.
    output_image: Optional[Image] (= None)
        Output label image fitlered.
    min_size: float (= 100)
        Smallest size object allowed.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_excludeLabelsOutsideSizeRange
    """
    return clic._remove_small_labels(device, input_image, output_image, float(min_size))

@plugin_function(categories=["label processing", "in assistant"])
def exclude_small_labels(
    input_image: Image,
    output_image: Optional[Image] =None,
    max_size: float =100,
    device: Optional[Device] =None
) -> Image:
    """Removes labels from a label map which are below a given maximum size.

    Parameters
    ----------
    input_image: Image 
        Label image to filter.
    output_image: Optional[Image] (= None)
        Output label image fitlered.
    max_size: float (= 100)
        Largest size object to exclude.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_excludeLabelsOutsideSizeRange
    """
    return clic._exclude_small_labels(device, input_image, output_image, float(max_size))

@plugin_function(categories=["label processing", "in assistant", "bia-bob-suggestion"])
def remove_large_labels(
    input_image: Image,
    output_image: Optional[Image] =None,
    max_size: float =100,
    device: Optional[Device] =None
) -> Image:
    """Removes labelled objects bigger than a given size (in pixels) from a label map.

    Parameters
    ----------
    input_image: Image 
        Label image to filter.
    output_image: Optional[Image] (= None)
        Output label image fitlered.
    max_size: float (= 100)
        Biggest size object allowed.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_excludeLabelsOutsideSizeRange
    """
    return clic._remove_large_labels(device, input_image, output_image, float(max_size))

@plugin_function(categories=["label processing", "in assistant"])
def exclude_large_labels(
    input_image: Image,
    output_image: Optional[Image] =None,
    min_size: float =100,
    device: Optional[Device] =None
) -> Image:
    """Removes labels from a label map which are higher a given minimum size.

    Parameters
    ----------
    input_image: Image 
        Label image to filter.
    output_image: Optional[Image] (= None)
        Output label image fitlered.
    min_size: float (= 100)
        Smallest size object to keep.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_excludeLabelsOutsideSizeRange
    """
    return clic._exclude_large_labels(device, input_image, output_image, float(min_size))


import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_indexing():
    data = np.asarray([[1.0, 2.0, 3.0], [1.0, 2.0, 3.0]]).astype(np.float32)

    cl_data = cle.push(data)

    assert cl_data[0, 0] == 1
    assert cl_data[0, 1] == 2



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_paste():
    test1 = cle.push(
        np.asarray([[0, 0, 0, 1], [0, 0, 3, 1], [0, 0, 3, 1], [1, 1, 1, 1]])
    )
    test2 = cle.push(
        np.asarray(
            [
                [1, 2],
            ]
        )
    )

    reference = cle.push(
        np.asarray([[0, 0, 0, 1], [0, 0, 3, 1], [0, 1, 2, 1], [1, 1, 1, 1]])
    )

    result = cle.create(test1)
    cle.copy(test1, result)
    cle.paste(test2, result, 1, 2, 0)

    a = cle.pull(result)
    b = cle.pull(reference)
    print(a)

    assert np.array_equal(a, b)



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_binary_edge_detection():
    test1 = cle.push(
        np.asarray(
            [
                [0, 0, 0, 0, 0, 0],
                [0, 1, 1, 1, 1, 0],
                [0, 1, 1, 1, 1, 0],
                [0, 1, 1, 1, 1, 0],
                [0, 1, 1, 1, 1, 0],
                [0, 0, 0, 0, 0, 0],
            ]
        )
    )

    test2 = cle.create_like(test1)
    cle.binary_edge_detection(test1, test2)

    print(test2)
    a = cle.pull(test2)
    assert np.min(a) == 0
    assert np.max(a) == 1
    assert np.mean(a) - 12 / 36 < 0.001



# This file is for code interroperability with other libraries. It is not expected to follow major
# conventions of pyclesperanto, instead it should be as simple as possible to use in order to make
# it easy to use pyclesperanto in other libraries and frameworks.
#
# The following code and behaviour is expected:
# Import aliases
# Function aliases or re-implementations
# Deprecated functions
# Deprecated aliases
#

import warnings
from typing import Optional

import numpy as np

from ._array import Image

# pyclesperanto_prototype aliases
from ._core import info as cl_info
from ._core import list_available_devices as available_device_names
from ._core import wait_for_kernel_to_finish

# numpy operations aliases
from ._memory import push as asarray

# scikit-image aliases
from ._tier5 import connected_components_labeling as label
from ._tier7 import affine_transform as _tier7_affine_transform

# enforce deprecation warnings display
warnings.filterwarnings("always", category=DeprecationWarning, module=__name__)


def affine_transform(
    input_image: Image,
    output_image: Image = None,
    transform_matrix: Optional[list] = None,
    interpolate: bool = False,
    resize: bool = False,
    transform: Optional[np.ndarray] = None,
    linear_interpolation: Optional[bool] = None,
    auto_size: Optional[bool] = None,
) -> Image:

    if transform is not None:
        warnings.warn(
            "affine_transform : 'transform_matrix' parameter is deprecated. Please use 'transform' instead.",
            DeprecationWarning,
        )
        transform_matrix = transform.ravel().tolist()

    if linear_interpolation is not None:
        warnings.warn(
            "affine_transform : 'linear_interpolation' parameter is deprecated. Please use 'interpolate' instead.",
            DeprecationWarning,
        )
        interpolate = linear_interpolation

    if auto_size is not None:
        warnings.warn(
            "affine_transform : 'auto_size' parameter is deprecated. Please use 'resize' instead.",
            DeprecationWarning,
        )
        resize = auto_size

    if not isinstance(transform_matrix, list) and isinstance(
        transform_matrix, np.ndarray
    ):
        transform_matrix = transform_matrix.ravel().tolist()

    return _tier7_affine_transform(
        input_image=input_image,
        output_image=output_image,
        transform_matrix=transform_matrix,
        interpolate=interpolate,
        resize=resize,
    )


def set_wait_for_kernel_finish(bool=True):
    warnings.warn(
        "set_wait_for_kernel_finish : This method is deprecated. Consider using wait_for_kernel_to_finish() instead.",
        DeprecationWarning,
    )
    wait_for_kernel_to_finish(bool)


def clip(a, a_min, a_max, out=None):
    from ._tier2 import clip

    a = asarray(a)
    if out:
        out = asarray(out)
    return clip(
        input_image=a,
        output_image=out,
        min_intensity=a_min,
        max_intensity=a_max,
        device=a.device,
    )


def mod(x1, x2, out=None):
    from ._tier1 import modulo_images

    x1 = asarray(x1)
    if out:
        out = asarray(out)
    return modulo_images(input_image0=x1, input_image1=x2, device=x1.device)


def sqrt(x, out=None):
    from ._tier1 import square_root

    x = asarray(x)
    if out:
        out = asarray(out)
    return square_root(input_image=x, output_image=out, device=x.device)


def cbrt(x, out=None):
    from ._tier1 import cubic_root

    x = asarray(x)
    if out:
        out = asarray(out)
    return cubic_root(input_image=x, output_image=out, device=x.device)


def power(x1, x2, out=None):
    x1 = asarray(x1)
    if out:
        out = asarray(out)

    # test if x2 is a scalar
    if np.isscalar(x2):
        from ._tier1 import power

        return power(input_image=x1, scalar=x2, output_image=out, device=x1.device)
    else:
        from ._tier1 import power_images

        x2 = asarray(x2)
        return power_images(
            input_image0=x1, input_image1=x2, output_image=out, device=x1.device
        )


def fabs(x, out=None):
    from ._memory import create
    from ._tier1 import absolute

    x = asarray(x)
    if out:
        out = asarray(out)
    else:
        out = create(x.shape, dtype=float, mtype=x.mtype, device=x.device)
    return absolute(input_image=x, output_image=out, device=x.device)



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_divide_scalar_by_image():
    test1 = cle.push(np.asarray([[5, 5], [1, 1], [2, 2]]))

    reference = cle.push(np.asarray([[2, 2], [10, 10], [5, 5]]))
    # reference = cle.push(np.asarray([[0.5, 0.5], [0.1, 0.1], [0.2, 0.2]]))

    result = cle.create(test1, dtype=float)
    cle.divide_scalar_by_image(test1, result, 10)

    a = cle.pull(result)
    b = cle.pull(reference)

    print(a)
    print(b)

    assert np.array_equal(a, b)



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_binary_or():
    test = cle.push(np.asarray([[1, 0], [1, 0]]))

    test1 = cle.push(np.asarray([[1, 1], [0, 0]]))

    test2 = cle.create(test)
    cle.binary_or(test, test1, test2)

    print(test2)

    a = cle.pull(test2)
    assert np.min(a) == 0
    assert np.max(a) == 1
    assert np.mean(a) == 0.75



import numpy as np

import pyclesperanto as cle

absolute_ocl = """
__constant sampler_t sampler = CLK_NORMALIZED_COORDS_FALSE | CLK_ADDRESS_CLAMP_TO_EDGE | CLK_FILTER_NEAREST;

__kernel void absolute(
    IMAGE_src_TYPE  src,
    IMAGE_dst_TYPE  dst
)
{
  const int x = get_global_id(0);
  const int y = get_global_id(1);
  const int z = get_global_id(2);

  IMAGE_src_PIXEL_TYPE value = READ_IMAGE(src, sampler, POS_src_INSTANCE(x,y,z,0)).x;
  if ( value < 0 ) {
    value = -1 * value;
  }

  WRITE_IMAGE(dst, POS_dst_INSTANCE(x,y,z,0), CONVERT_dst_PIXEL_TYPE(value));
}
"""

add_native_ocl = """
__kernel void add_arrays(__global const float* a, __global const float* b, __global float* c, const unsigned int n) {
    int id = get_global_id(0);
    if (id < n) {
        c[id] = a[id] + b[id];
    }
}
"""


def test_execute_native_from_file():
    input1 = cle.push(np.ones(10).astype(float))
    input2 = cle.push(np.ones(10).astype(float) * 2)
    output = cle.create(input1)

    param = {"a": input1, "b": input2, "c": output, "n": int(np.prod(input1.shape))}
    cle.native_execute(
        anchor=__file__,
        kernel_source="_test_add_arrays.cl",
        kernel_name="add_arrays",
        global_size=input1.shape,
        local_size=(1, 1, 1),
        parameters=param,
    )

    print(output)

    a = cle.pull(output)
    assert np.min(a) == 3
    assert np.max(a) == 3
    assert np.mean(a) == 3


def test_execute_absolute_from_file():
    input = cle.push(np.asarray([[1, -1], [1, -1]]).astype(float))
    output = cle.create(input)

    param = {"src": input, "dst": output}
    cle.execute(
        anchor=__file__,
        kernel_source="_test_absolute.cl",
        kernel_name="absolute",
        global_size=input.shape,
        parameters=param,
    )

    print(output)

    a = cle.pull(output)
    assert np.min(a) == 1
    assert np.max(a) == 1
    assert np.mean(a) == 1


def test_execute_native():
    input1 = cle.push(np.ones(10).astype(float))
    input2 = cle.push(np.ones(10).astype(float) * 2)
    output = cle.create(input1)

    param = {"a": input1, "b": input2, "c": output, "n": int(np.prod(input1.shape))}
    cle.native_execute(
        kernel_source=add_native_ocl,
        kernel_name="add_arrays",
        global_size=input1.shape,
        local_size=(1, 1, 1),
        parameters=param,
    )

    print(output)

    a = cle.pull(output)
    assert np.min(a) == 3
    assert np.max(a) == 3
    assert np.mean(a) == 3


def test_execute_absolute():
    input = cle.push(np.asarray([[1, -1], [1, -1]]).astype(float))
    output = cle.create(input)

    param = {"src": input, "dst": output}
    cle.execute(
        kernel_source=absolute_ocl,
        kernel_name="absolute",
        global_size=input.shape,
        parameters=param,
    )

    print(output)

    a = cle.pull(output)
    assert np.min(a) == 1
    assert np.max(a) == 1
    assert np.mean(a) == 1



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_detect_label_edges():
    test = cle.push(
        np.asarray(
            [
                [1, 1, 2, 2, 2],
                [1, 1, 2, 2, 2],
                [3, 3, 3, 2, 2],
                [3, 3, 3, 2, 0],
                [3, 3, 3, 0, 0],
            ]
        )
    )

    result = cle.create(test)
    cle.detect_label_edges(test, result)

    reference = cle.push(
        np.asarray(
            [
                [0, 1, 1, 0, 0],
                [1, 1, 1, 0, 0],
                [1, 1, 1, 1, 1],
                [0, 0, 1, 1, 1],
                [0, 0, 1, 1, 0],
            ]
        )
    )

    print(result)

    a = cle.pull(result)
    b = cle.pull(reference)

    print(b)

    assert np.array_equal(a, b)



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_replace_intensity():
    test1 = cle.push(
        np.asarray(
            [
                [0, 0, 0, 0, 0],
                [0, 1, 2, 3, 0],
                [0, 2, 3, 4, 0],
                [0, 4, 4, 5, 0],
                [0, 0, 0, 0, 0],
            ]
        )
    )

    reference = cle.push(
        np.asarray(
            [
                [0, 0, 0, 0, 0],
                [0, 1, 8, 3, 0],
                [0, 8, 3, 4, 0],
                [0, 4, 4, 5, 0],
                [0, 0, 0, 0, 0],
            ]
        )
    )

    result = cle.create(test1)
    cle.replace_value(test1, result, 2, 8)

    print(result)

    a = cle.pull(result)
    b = cle.pull(reference)
    assert np.allclose(a, b, 0.001)



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_crop():
    test1 = cle.push(
        np.asarray([[0, 0, 0, 1], [0, 0, 3, 1], [0, 0, 3, 1], [1, 1, 1, 1]])
    )

    reference = cle.push(np.asarray([[0, 0, 0], [0, 0, 3], [0, 0, 3]]))

    result = cle.create_like(reference)
    cle.crop(test1, result, 0, 0)

    a = cle.pull(result)
    b = cle.pull(reference)

    print(a)
    assert np.array_equal(a, b)


def test_crop_2d():
    input_image = cle.create([100, 100])
    print(input_image.shape)

    output_image = cle.crop(input_image, width=10, height=10)

    print(output_image.shape)
    assert len(output_image.shape) == 2
    assert output_image.shape[0] == 10
    assert output_image.shape[1] == 10


def test_crop_3d():
    input_image = cle.create([100, 100, 100])
    print(input_image.shape)

    output_image = cle.crop(input_image, width=10, height=10, depth=10)

    print(output_image.shape)
    assert len(output_image.shape) == 3
    assert output_image.shape[0] == 10
    assert output_image.shape[1] == 10
    assert output_image.shape[2] == 10



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_hessian_eigenvalues_2d():
    test = np.asarray([[1, -1], [1, -1]])

    reference_small_hessian_eigenvalue = np.asarray([[-2, 0], [-2, 0]])

    reference_large_hessian_eigenvalue = np.asarray([[0, 2], [0, 2]])

    small_hessian_eigenvalue, large_hessian_eigenvalue = cle.hessian_eigenvalues(test)

    print(small_hessian_eigenvalue)
    print(large_hessian_eigenvalue)

    assert np.allclose(reference_small_hessian_eigenvalue, small_hessian_eigenvalue)
    assert np.allclose(reference_large_hessian_eigenvalue, large_hessian_eigenvalue)


def test_hessian_eigenvalues_3d():
    test = np.asarray(
        [
            [[1, -1], [1, -1]],
            [[2, -2], [2, -2]],
        ]
    )

    reference_small_hessian_eigenvalue = np.asarray(
        [
            [[-2.1, -1.1], [-2.1, -1.1]],
            [[-4.1, 0], [-4.1, 0]],
        ]
    )

    reference_middle_hessian_eigenvalue = np.asarray(
        [
            [[0, 0], [0, 0]],
            [[-0.9, 0.9], [-0.9, 0.9]],
        ]
    )

    reference_large_hessian_eigenvalue = np.asarray(
        [
            [[1.1, 2.1], [1.1, 2.1]],
            [[0, 4.1], [0, 4.1]],
        ]
    )

    (
        small_hessian_eigenvalue,
        middle_hessian_eigenvalue,
        large_hessian_eigenvalue,
    ) = cle.hessian_eigenvalues(test)

    print(small_hessian_eigenvalue)
    print(reference_small_hessian_eigenvalue)
    print()
    print(middle_hessian_eigenvalue)
    print(reference_middle_hessian_eigenvalue)
    print()
    print(large_hessian_eigenvalue)
    print(reference_large_hessian_eigenvalue)

    assert np.allclose(
        reference_small_hessian_eigenvalue, small_hessian_eigenvalue, atol=0.1
    )
    assert np.allclose(
        reference_middle_hessian_eigenvalue, middle_hessian_eigenvalue, atol=0.1
    )
    assert np.allclose(
        reference_large_hessian_eigenvalue, large_hessian_eigenvalue, atol=0.1
    )



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_generate_binary_overlap_matrix_2d():
    gpu_input1 = cle.push(
        np.asarray(
            [
                [1, 1, 0, 0, 0],
                [1, 1, 0, 3, 0],
                [0, 2, 2, 3, 0],
                [0, 2, 2, 0, 0],
                [0, 0, 0, 0, 4],
            ]
        )
    )

    gpu_input2 = cle.push(
        np.asarray(
            [
                [1, 1, 2, 2, 2],
                [1, 1, 2, 2, 2],
                [1, 0, 0, 2, 2],
                [1, 0, 0, 2, 2],
                [1, 2, 2, 2, 2],
            ]
        )
    )

    gpu_reference = cle.push(
        np.asarray([[0, 1, 1], [0, 1, 0], [1, 0, 0], [0, 0, 1], [0, 0, 1]]).T
    )

    gpu_binary_overlap_matrix = cle.generate_binary_overlap_matrix(
        gpu_input1, gpu_input2
    )

    a = cle.pull(gpu_binary_overlap_matrix)
    b = cle.pull(gpu_reference)

    print(a)
    print(b)

    assert np.allclose(a, b, 0.01)


def test_generate_binary_overlap_matrix_3d():
    gpu_input1 = cle.push(
        np.asarray(
            [
                [
                    [1, 1, 0, 0, 0],
                    [1, 1, 0, 3, 0],
                ],
                [[0, 2, 2, 0, 0], [0, 0, 0, 0, 4]],
            ]
        )
    )

    gpu_input2 = cle.push(
        np.asarray(
            [
                [
                    [1, 1, 2, 2, 2],
                    [1, 1, 2, 2, 2],
                ],
                [[1, 0, 0, 2, 2], [1, 2, 2, 2, 2]],
            ]
        )
    )

    gpu_reference = cle.push(
        np.asarray([[0, 1, 1], [0, 1, 0], [1, 0, 0], [0, 0, 1], [0, 0, 1]]).T
    )

    gpu_binary_overlap_matrix = cle.generate_binary_overlap_matrix(
        gpu_input1, gpu_input2
    )

    a = cle.pull(gpu_binary_overlap_matrix)
    b = cle.pull(gpu_reference)

    print(a)
    print(b)

    assert np.allclose(a, b, 0.01)



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_subtract_gaussian_background():
    test1 = cle.push(np.asarray([[0, 0], [1, 1], [2, 2]]))

    reference = cle.difference_of_gaussian(
        test1, sigma1_x=0, sigma1_y=0, sigma1_z=0, sigma2_x=5, sigma2_y=5, sigma2_z=5
    )

    result = cle.create(test1, dtype=float)
    cle.subtract_gaussian_background(test1, result, 5, 5, 5)

    a = cle.pull(result)
    b = cle.pull(reference)

    print(a)

    assert np.array_equal(a, b)



import pyclesperanto as cle
import numpy as np


def test_extension_ratio_map_2d():

    labels = cle.push(np.asarray([[1, 1, 2], [1, 0, 2], [3, 3, 0]]))

    reference = cle.push(
        np.asarray([[1.1396203, 1.1396203, 1], [1.1396203, 0, 1], [1, 1, 0]])
    )

    result = cle.extension_ratio_map(labels)

    a = cle.pull(result)
    b = cle.pull(reference)

    print(a)
    print(b)

    assert np.allclose(a, b, 0.001)


def test_extension_ratio_map_3d():

    labels = cle.push(
        np.asarray(
            [
                [
                    [1, 1, 2],
                ],
                [
                    [1, 0, 2],
                ],
                [
                    [3, 3, 0],
                ],
            ]
        )
    )

    reference = cle.push(
        np.asarray(
            [
                [
                    [1.1396203, 1.1396203, 1],
                ],
                [
                    [1.1396203, 0, 1],
                ],
                [
                    [1, 1, 0],
                ],
            ]
        )
    )

    result = cle.extension_ratio_map(labels)

    a = cle.pull(result)
    b = cle.pull(reference)

    print(a)
    print(b)

    assert np.allclose(a, b, 0.001)



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_range1():
    image = cle.push(np.asarray([[0, 1, 2], [2, 3, 4], [5, 6, 7]]))

    reference = cle.push(np.asarray([[3, 4], [6, 7]]))

    crop = image[1:, 1:]

    print(reference)
    print(crop)

    assert np.array_equal(crop, reference)


def test_range2():
    image = cle.push(np.asarray([[0, 1, 2], [2, 3, 4], [5, 6, 7]]))

    reference = cle.push(np.asarray([[0, 1], [2, 3]]))

    crop = image[:2, :2]

    print(reference)
    print(crop)

    assert np.array_equal(crop, reference)


def test_range3():
    image = cle.push(np.asarray([[[0, 1, 2], [2, 3, 4], [5, 6, 7]]]))

    reference = cle.push(np.asarray([[[3, 4], [6, 7]]]))

    crop = image[:, 1:, 1:]

    print(reference)
    print(crop)

    assert np.array_equal(crop, reference)


def test_range4():
    image = cle.push(np.asarray([[[0, 1, 2], [2, 3, 4], [5, 6, 7]]]))

    reference = cle.push(np.asarray([[[0, 1], [2, 3]]]))

    crop = image[:, :2, :2]

    print(reference)
    print(crop)

    assert np.array_equal(crop, reference)


def test_range5():
    image = cle.push(np.random.random((10, 20, 30)))

    crop = image[:5]

    assert crop.shape[0] == 5
    assert crop.shape[1] == 20
    assert crop.shape[2] == 30


def test_range6():
    image = cle.push(np.random.random((10, 20, 30)))

    crop = image[5]

    assert crop.shape[0] == 20
    assert crop.shape[1] == 30
    assert len(crop.shape) == 2


def test_range7():
    image = cle.push(np.random.random((10, 20, 30)))

    crop = image[5, :]

    assert crop.shape[0] == 20
    assert crop.shape[1] == 30
    assert len(crop.shape) == 2


def test_range8():
    image = cle.push(np.random.random((10, 20, 30)))

    crop = image[:, :, 5]

    assert crop.shape[0] == 10
    assert crop.shape[1] == 20
    assert len(crop.shape) == 2


def test_range9():
    image = cle.push(np.random.random((10, 20, 30)))

    crop = image[:, 5]

    assert crop.shape[0] == 10
    assert crop.shape[1] == 30
    assert len(crop.shape) == 2


def test_range_against_numpy_1():
    input = np.random.random((2, 3, 4))
    input_gpu = cle.push(input)

    reference = input[1, :, :]
    result = input_gpu[1, :, :]

    print(reference)
    print(result)

    assert np.allclose(reference, result, 0.0001)


def test_range_against_numpy_2():
    input = np.random.random((2, 3, 4))
    input_gpu = cle.push(input)

    reference = input[:, 1, :]
    result = input_gpu[:, 1, :]

    print(reference)
    print(result)

    assert np.allclose(reference, result, 0.0001)


def test_range_against_numpy_3():
    input = np.random.random((2, 3, 4))
    input_gpu = cle.push(input)

    reference = input[:, :, 1]
    result = input_gpu[:, :, 1]

    print(reference)
    print(result)

    assert np.allclose(reference, result, 0.0001)


def test_range_against_numpy_4():
    input = np.random.random((2, 3, 4))
    input_gpu = cle.push(input)

    reference = input[1, :]
    result = input_gpu[1, :]

    print(reference)
    print(result)

    assert np.allclose(reference, result, 0.0001)


def test_range_against_numpy_5():
    input = np.random.random((2, 3, 4))
    input_gpu = cle.push(input)

    reference = input[1]
    result = input_gpu[1]

    print(reference)
    print(result)

    assert np.allclose(reference, result, 0.0001)


def test_range_against_numpy_6():
    input = np.random.random((10, 20, 30))
    input_gpu = cle.push(input)

    reference = input[2:5, :, :]
    result = input_gpu[2:5, :, :]

    print(reference)
    print(result)

    assert np.allclose(reference, result, 0.0001)


def test_range_against_numpy_7():
    input = np.random.random((10, 20, 30))
    input_gpu = cle.push(input)

    reference = input[:, 10:15, :]
    result = input_gpu[:, 10:15, :]

    print(reference)
    print(result)

    assert np.allclose(reference, result, 0.0001)


def test_range_against_numpy_8():
    input = np.random.random((10, 20, 30))
    input_gpu = cle.push(input)

    reference = input[:, :, 10:15]
    result = input_gpu[:, :, 10:15]

    print(reference)
    print(result)

    assert np.allclose(reference, result, 0.0001)


def test_range_against_numpy_9():
    input = np.random.random((10, 20, 30))
    input_gpu = cle.push(input)

    reference = input[:5, :, :]
    result = input_gpu[:5, :, :]

    print(reference)
    print(result)

    assert np.allclose(reference, result, 0.0001)


def test_range_against_numpy_10():
    input = np.random.random((10, 20, 30))
    input_gpu = cle.push(input)

    reference = input[:, :5, :]
    result = input_gpu[:, :5, :]

    print(reference)
    print(result)

    assert np.allclose(reference, result, 0.0001)


def test_range_against_numpy_11():
    input = np.random.random((10, 20, 30))
    input_gpu = cle.push(input)

    reference = input[:, :, :5]
    result = input_gpu[:, :, :5]

    print(reference)
    print(result)

    assert np.allclose(reference, result, 0.0001)


def test_range_against_numpy_12():
    input = np.random.random((10, 20, 30))
    input_gpu = cle.push(input)

    reference = input[5:, :, :]
    result = input_gpu[5:, :, :]

    print(reference)
    print(result)

    assert np.allclose(reference, result, 0.0001)


def test_range_against_numpy_14():
    input = np.random.random((10, 20, 30))
    input_gpu = cle.push(input)

    reference = input[:, 5:, :]
    result = input_gpu[:, 5:, :]

    print(reference)
    print(result)

    assert np.allclose(reference, result, 0.0001)


def test_range_against_numpy_15():
    input = np.random.random((10, 20, 30))
    input_gpu = cle.push(input)

    reference = input[:, :, 5:]
    result = input_gpu[:, :, 5:]

    print(reference)
    print(result)

    assert np.allclose(reference, result, 0.0001)


def test_range_against_numpy_16():
    input = np.random.random((10, 20, 30))
    input_gpu = cle.push(input)

    reference = input[:, :, np.int32(5)]
    result = input_gpu[:, :, np.int32(5)]

    print(reference)
    print(result)

    assert np.allclose(reference, result, 0.0001)


def test_type_1():
    input = np.random.random((10, 20, 30)).astype(np.uint16)
    input_gpu = cle.push(input).astype(np.uint16)

    reference = input[:, :, np.int32(5)]
    result = input_gpu[:, :, np.int32(5)]

    print(reference)
    print(result)

    assert reference.dtype == result.dtype


def test_types_2():
    input = np.random.random((10, 20, 30)).astype(np.uint16)
    input_gpu = cle.push(input).astype(np.uint16)

    reference = input[:, :, np.int64(5)]
    result = input_gpu[:, :, np.int64(5)]

    print(reference)
    print(result)

    assert np.allclose(reference, result, 0.0001)


def test_negative_step_2d():
    numbers = np.reshape(np.asarray([[i] for i in range(0, 20)]), (2, 10))

    cle_numbers = cle.asarray(numbers)

    assert cle.array_equal(numbers[::-1], cle_numbers[::-1])
    assert cle.array_equal(numbers[1::-1], cle_numbers[1::-1])
    # assert cle.array_equal(numbers[:5:-1], cle_numbers[:5:-1])
    # assert cle.array_equal(numbers[5:1:-1], cle_numbers[5:1:-1])
    # assert cle.array_equal(numbers[100:], cle_numbers[100:])
    assert cle.array_equal(numbers[100::-1], cle_numbers[100::-1])
    assert cle.array_equal(numbers[100:-50:-1], cle_numbers[100:-50:-1])

    assert cle.array_equal(numbers[::, ::-1], cle_numbers[::, ::-1])
    assert cle.array_equal(numbers[::, 1::-1], cle_numbers[::, 1::-1])
    # assert cle.array_equal(numbers[::, :5:-1], cle_numbers[::, :5:-1])
    # assert cle.array_equal(numbers[::, 5:1:-1], cle_numbers[::, 5:1:-1])
    # assert cle.array_equal(numbers[::, 100:], cle_numbers[::, 100:])
    assert cle.array_equal(numbers[::, 100::-1], cle_numbers[::, 100::-1])
    assert cle.array_equal(numbers[::, 100:-50:-1], cle_numbers[::, 100:-50:-1])


def test_negative_step_3d():
    numbers = np.reshape(np.asarray([[i] for i in range(0, 60)]), (3, 4, 5))

    cle_numbers = cle.asarray(numbers)

    assert cle.array_equal(numbers[::-1], cle_numbers[::-1])
    assert cle.array_equal(numbers[1::-1], cle_numbers[1::-1])
    # assert cle.array_equal(numbers[:5:-1], cle_numbers[:5:-1])
    # assert cle.array_equal(numbers[5:1:-1], cle_numbers[5:1:-1])
    # assert cle.array_equal(numbers[100:], cle_numbers[100:])
    assert cle.array_equal(numbers[100::-1], cle_numbers[100::-1])
    assert cle.array_equal(numbers[100:-50:-1], cle_numbers[100:-50:-1])

    assert cle.array_equal(numbers[::, ::-1], cle_numbers[::, ::-1])
    assert cle.array_equal(numbers[::, 1::-1], cle_numbers[::, 1::-1])
    # assert cle.array_equal(numbers[::, :5:-1], cle_numbers[::, :5:-1])
    # assert cle.array_equal(numbers[::, 5:1:-1], cle_numbers[::, 5:1:-1])
    # assert cle.array_equal(numbers[::, 100:], cle_numbers[::, 100:])
    assert cle.array_equal(numbers[::, 100::-1], cle_numbers[::, 100::-1])
    assert cle.array_equal(numbers[::, 100:-50:-1], cle_numbers[::, 100:-50:-1])

    assert cle.array_equal(numbers[::, ::, ::-1], cle_numbers[::, ::, ::-1])
    assert cle.array_equal(numbers[::, ::, 1::-1], cle_numbers[::, ::, 1::-1])
    # assert cle.array_equal(numbers[::, ::, :5:-1], cle_numbers[::, ::, :5:-1])
    # assert cle.array_equal(numbers[::, ::, 5:1:-1], cle_numbers[::, ::, 5:1:-1])
    # assert cle.array_equal(numbers[::, ::, 100:], cle_numbers[::, ::, 100:])
    assert cle.array_equal(numbers[::, ::, 100::-1], cle_numbers[::, ::, 100::-1])
    assert cle.array_equal(numbers[::, ::, 100:-50:-1], cle_numbers[::, ::, 100:-50:-1])



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_undefined_to_zero():
    test1 = cle.push(
        np.asarray(
            [
                [0, 0, 0, 0, 0],
                [0, 1, 1, 0, 0],
                [0, 1, 1, 0, 0],
                [0, 1, 1, 0, 0],
                [0, 0, 0, 0, 0],
            ]
        ).astype(float)
    )

    test2 = cle.push(
        np.asarray(
            [
                [0, 0, 0, 0, 0],
                [0, 0, 1, 1, 0],
                [0, 0, 1, 1, 0],
                [0, 0, 1, 1, 0],
                [0, 0, 0, 0, 0],
            ]
        ).astype(float)
    )

    reference = cle.push(
        np.asarray(
            [
                [0, 0, 0, 0, 0],
                [0, 0, 1, 0, 0],
                [0, 0, 1, 0, 0],
                [0, 0, 1, 0, 0],
                [0, 0, 0, 0, 0],
            ]
        ).astype(float)
    )

    divided = cle.divide_images(test1, test2)

    divided_no_nan = cle.undefined_to_zero(divided)

    print(divided)
    print(divided_no_nan)

    a = cle.pull(divided_no_nan)
    b = cle.pull(reference)

    print(a)
    print(b)

    assert np.array_equal(a, b)



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")

input1 = np.asarray([1, 2, 3])
input2 = np.asarray([4, 5, 6])


def test_add_images():
    reference = np.asarray([5, 7, 9])
    output = cle.add_images(input1, input2, None)
    result = cle.pull(output)

    assert np.array_equal(result, reference)



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_detect_minima_box():
    gpu_input = cle.push(
        np.asarray(
            [
                [6, 6, 6, 6, 6],
                [5, 6, 6, 4, 6],
                [5, 4, 5, 3, 4],
                [6, 5, 6, 4, 6],
                [6, 6, 6, 6, 6],
            ]
        )
    )
    gpu_output = cle.create_like(gpu_input)

    gpu_reference = cle.push(
        np.asarray(
            [
                [0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0],
                [0, 1, 0, 1, 0],
                [0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0],
            ]
        )
    )

    cle.detect_minima(gpu_input, gpu_output)

    a = cle.pull(gpu_output)
    b = cle.pull(gpu_reference)

    print(a)
    print(b)

    assert np.array_equal(a, b)



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_modulo_images():
    test = [[2, 5], [2, 3]]
    test_div = [[2, 2], [2, 2]]
    reference = [[0, 1], [0, 1]]

    result = cle.modulo_images(test, test_div)

    assert np.array_equal(result, reference)



import warnings
from typing import Optional, Tuple

import numpy as np

from ._array import Array, Image
from ._core import Device, get_device


def create(
    dim,
    dtype: Optional[type] = None,
    mtype: Optional[str] = None,
    device: Optional[Device] = None,
) -> Array:
    """Create a new image on the device.

    Parameters
    ----------
    shape : Tuple[int, ...]
        Shape of the image in (z,y,x)
    dtype : type, optional
        Data type of the image (np.int8, np.float32, etc.), np.float32 by default if None
    mtype : str, optional
        Memory type of the image (buffer, image), buffer by default if None
    device : Device, optional
        Device on which the image is created, current device by default if None

    Returns
    -------
    Array
        Created an empty Array on the device
    """
    if isinstance(dim, Array):
        device = device if device else dim.device
        mtype = mtype if mtype else dim.mtype
    if isinstance(dim, (np.ndarray, Array)):
        dtype = dtype if dtype else dim.dtype
        dim = dim.shape
    else:
        dim = tuple(dim)

    device = device if device else get_device()
    dtype = dtype if dtype else float
    mtype = mtype if mtype else "buffer"

    return Array.empty(dim, dtype=dtype, mtype=mtype, device=device)


def create_like(
    array: Image,
    dtype: Optional[type] = None,
    mtype: Optional[str] = None,
    device: Optional[Device] = None,
) -> Array:
    """Create a new image on the device with the same shape and dtype as the input image.

    Parameters
    ----------
    array : Image
        Input image
    dtype : type, optional
        Data type of the image (np.int8, np.float32, etc.), np.float32 by default if None
    mtype : str, optional
        Memory type of the image (buffer, image), buffer by default if None
    device : Device, optional
        Device on which the image is created, current device by default if None

    Returns
    -------
    Array
        Created an empty Array on the device
    """
    return create(array.shape, dtype, mtype, device)


def push(
    array: Image,
    dtype: Optional[type] = None,
    mtype: Optional[str] = None,
    device: Optional[Device] = None,
) -> Array:
    """Create a new image on the device and push the input image into it.

    Parameters
    ----------
    array : Image
        Input image
    dtype : type, optional
        If provided, the input image is cast to the given dtype before being
        pushed to the device. Examples are `np.int8`, `np.float32`, etc.
        By default, no casting is performed and the dtype of the pushed
        image will match the dtype of the input image.
    mtype : str, optional
        Memory type of the image (buffer, image), buffer by default if None
    device : Device, optional
        Device on which the image is created, current device by default if None

    Returns
    -------
    Array
        Created Array on the device with the input image data
    """
    if isinstance(array, Array):
        return array

    if not isinstance(array, np.ndarray):
        array = np.asarray(array)

    dtype = dtype if dtype else array.dtype
    return create(array.shape, dtype, mtype, device).set(array)


def pull(array: Image) -> np.ndarray:
    """Pull the input image from the device to the host.

    Parameters
    ----------
    array : Image
        Input image

    Returns
    -------
    np.ndarray
        Image data
    """
    if isinstance(array, Array):
        return array.get()
    return array



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_jaccard_index_2d():
    test1 = cle.push(np.asarray([[0, 0, 0, 0, 0], [0, 1, 1, 0, 0]]))

    test2 = cle.push(np.asarray([[0, 1, 1, 0, 0], [0, 1, 1, 0, 0]]))

    j = cle.jaccard_index(test1, test2)

    assert j == 0.5


def test_accard_index_3d():
    test1 = cle.push(np.asarray([[[0, 0, 0], [0, 0, 0]], [[0, 1, 1], [0, 1, 0]]]))

    test2 = cle.push(np.asarray([[[0, 1, 1], [0, 1, 0]], [[0, 1, 1], [0, 1, 0]]]))

    j = cle.jaccard_index(test1, test2)

    assert j == 0.5



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_maximum_y_projection():
    test1 = cle.push(
        np.asarray(
            [
                [
                    [1, 0, 0, 0, 9],
                    [0, 2, 0, 8, 0],
                    [3, 0, 1, 0, 10],
                    [0, 4, 0, 7, 0],
                    [5, 0, 6, 0, 10],
                ],
                [
                    [0, 2, 0, 8, 0],
                    [1, 0, 0, 0, 9],
                    [3, 0, 1, 0, 10],
                    [0, 4, 0, 7, 0],
                    [5, 0, 6, 0, 10],
                ],
                [
                    [0, 2, 0, 8, 0],
                    [3, 0, 1, 0, 10],
                    [0, 4, 0, 7, 0],
                    [1, 0, 0, 0, 9],
                    [5, 0, 6, 0, 10],
                ],
                [
                    [0, 2, 0, 8, 0],
                    [1, 0, 0, 0, 9],
                    [0, 4, 0, 7, 0],
                    [3, 0, 1, 0, 10],
                    [5, 0, 6, 0, 10],
                ],
                [
                    [1, 0, 0, 0, 9],
                    [0, 4, 0, 7, 0],
                    [3, 0, 1, 0, 10],
                    [0, 2, 0, 8, 0],
                    [5, 0, 6, 0, 10],
                ],
            ]
        )
    )

    reference = cle.push(
        np.asarray(
            [
                [5, 4, 6, 8, 10],
                [5, 4, 6, 8, 10],
                [5, 4, 6, 8, 10],
                [5, 4, 6, 8, 10],
                [5, 4, 6, 8, 10],
            ]
        )
    )

    result = cle.create(reference)
    cle.maximum_y_projection(test1, result)

    a = cle.pull(result)
    b = cle.pull(reference)

    print(a)

    assert np.array_equal(a, b)


def test_maximum_y_projection2():
    test1 = cle.push(
        np.asarray(
            [
                [1, 0, 0, 0, 9],
                [0, 2, 0, 8, 0],
                [3, 0, 1, 0, 10],
            ]
        )
    )

    reference = cle.push(np.asarray([3, 2, 1, 8, 10]))

    result = cle.maximum_y_projection(test1)

    a = cle.pull(result)
    b = cle.pull(reference)

    print(a)

    assert np.array_equal(a, b)


def test_maximum_y_projection_against_numpy():
    from skimage.data import camera

    image = camera()

    max_cle = cle.maximum_y_projection(image)
    max_np = image.max(axis=0)

    print(max_cle.get())
    print(max_np)

    assert np.array_equal(max_cle.get(), max_np)


def test_maximum_y_projection_against_numpy_small():
    from skimage.data import camera

    image = camera()[0:2, 0:10]

    max_cle = cle.maximum_y_projection(image)
    max_np = image.max(axis=0)

    print(max_cle.get())
    print(max_np)

    assert np.array_equal(max_cle.get(), max_np)
